def msgfmt(i):
    return '"{0}": "{1}"'.format(
        i.msgid,
        i.msgstr,
    )


def open_po_file(fn):
    import polib

    try:
        return polib.pofile(str(fn))
    except Exception as e:
        raise TranslationException(
            str(e),
            [],
        )


def open_mo_file(fn):
    import polib

    try:
        return polib.mofile(str(fn))
    except Exception as e:
        raise TranslationException(
            str(e),
            [],
        )


class TranslationException(Exception):
    pass



"""A py.test plugin to check ``gettext`` ``po`` & ``mo`` files."""


def pytest_addoption(parser):
    group = parser.getgroup("general")
    group.addoption(
        '--translations',
        action='store_true',
        help="perform some checks on .mo and .po files"
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "translations: translation tests"
    )


def pytest_collect_file(path, parent):
    from .mo_files import MoFileItem
    from .po_files import PoFile

    config = parent.config
    if config.option.translations:
        if path.ext == '.mo':
            return MoFileItem.from_parent(
                fspath=path,
                parent=parent,
            )
        elif path.ext == '.po':
            return PoFile.from_parent(
                fspath=path,
                parent=parent
            )



# coding=utf8
import os

import polib
import pytest


pytest_plugins = "pytester",


class TestMo(object):
    @pytest.fixture
    def pomo(self, testdir):
        pofile = testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            msgid "car"
            msgstr "Auto"
            """
        )

        p, ext = os.path.splitext(str(pofile))
        mofile = p + ".mo"

        polib.pofile(str(pofile)).save_as_mofile(mofile)

        return str(pofile), str(mofile)

    def test_broken_file(self, testdir):
        testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            msgid "car"
            msgstr "Auto"
            """
        )
        testdir.makefile(
            '.mo',
            """
            asdflkaj sdlkfaj
            """,
        )

        result = testdir.runpytest("--translations")
        assert result.ret == 1
        result.stdout.fnmatch_lines([
            "*collected 5*",
            "*Invalid mo file*",
            "*1 failed*",
        ])

    def test_ok(self, testdir, pomo):
        result = testdir.runpytest("--translations", '-vvv')
        assert result.ret == 0
        result.stdout.fnmatch_lines([
            "*collected 5*",
            "*4 passed*",
        ])

    def test_missing_po(self, testdir, pomo):
        po, mo = pomo
        os.unlink(po)

        result = testdir.runpytest("--translations", '-vvv')
        assert result.ret == 1
        result.stdout.fnmatch_lines([
            "*collected 1*",
            "*corresponding .po file does not exist*",
            "*1 failed*",
        ])

    def test_entry_mismatch(self, testdir, pomo):
        po, mo = pomo
        os.unlink(po)

        testdir.makefile(
            '.po',
            """
            msgid ""
            msgstr ""
            "Language: de\n"

            #: asdf.py:111
            msgid "bike"
            msgstr "Fahrrad"
            """
        )

        result = testdir.runpytest("--translations", '-vvv')
        assert result.ret == 1
        result.stdout.fnmatch_lines([
            "*collected 5*",
            "*mo*file content does not match po*",
            "*bike*Fahrrad*",
        ])


class TestPo(object):
    def test_uses_argument(self, testdir):
        testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            msgid "car"
            msgstr "Auto"
            """
        )
        result = testdir.runpytest()
        result.stdout.fnmatch_lines([
            "*collected 0*",
        ])

    def test_broken_file(self, testdir):
        testdir.makefile(
            '.po',
            """
            asdflkaj sdlkfaj
            """
        )
        result = testdir.runpytest('--translations', '-vvv')
        assert result.ret == 1
        result.stdout.fnmatch_lines([
            "*Syntax error*",
        ])

    def test_valid(self, testdir):
        testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            msgid "car"
            msgstr "Auto"
            """
        )
        result = testdir.runpytest('--translations', '-vvv')
        assert result.ret == 0
        result.stdout.fnmatch_lines([
            "*collected 4*",
            "*3 passed*",
        ])

    def test_missing_translation(self, testdir):
        testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            msgid "car"
            msgstr ""
            """
        )
        result = testdir.runpytest('--translations', '-vvv')
        assert result.ret == 1
        result.stdout.fnmatch_lines([
            "*collected 3*",
            "*1 failed*2 passed*",
        ])

    def test_fuzzy(self, testdir):
        testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            #, fuzzy
            msgid "car"
            msgstr "Auto"
            """
        )
        result = testdir.runpytest('--translations', '-vvv')
        assert result.ret == 1
        result.stdout.fnmatch_lines([
            "*collected 3*",
            "*fuzzy*",
            "*1 failed*2 passed*",
        ])

    def test_obsolete(self, testdir):
        testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            #~ msgid "car"
            #~ msgstr "Auto"
            """
        )
        result = testdir.runpytest('--translations', '-vvv')
        assert result.ret == 1
        result.stdout.fnmatch_lines([
            "*collected 3*",
            "*obsolete*",
            "*1 failed*2 passed*",
        ])

    def test_all(self, testdir):
        testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            msgid "car2"
            msgstr ""

            #: asdf.py:111
            #, fuzzy
            msgid "car1"
            msgstr "Auto1"

            #: asdf.py:111
            #~ msgid "car"
            #~ msgstr "Auto"
            """
        )
        result = testdir.runpytest('--translations', '-vvv')
        assert result.ret == 1
        result.stdout.fnmatch_lines([
            "*collected 3*",
            "*untranslated*",
            "*fuzzy*",
            "*obsolete*",
            "*3 failed*",
        ])


class TestPoSpellcheck(object):
    def test_broken_file(self, testdir):
        testdir.makefile(
            '.po',
            """
            asdflkjasdf laskdjfasdf
            """
        )
        result = testdir.runpytest('--translations', '-vvv', '-r', 's')
        result.stdout.fnmatch_lines([
            "*collected 3*",
        ])

    def test_language_missing_in_po(self, testdir):
        testdir.makefile(
            '.po',
            """
            #: asdf.py:111
            msgid "meeting"
            msgstr "meeeting"
            """
        )
        result = testdir.runpytest('--translations', '-vvv', '-r', 's')
        result.stdout.fnmatch_lines([
            "*collected 4*",
            "SKIPPED * no language defined in PO file",
        ])

    def test_language_catalog_missing(self, testdir):
        testdir.makefile(
            '.po',
            """
            msgid ""
            msgstr ""
            "Language: hr\\n"

            #: asdf.py:111
            msgid "meeting"
            msgstr "meeeting"
            """
        )
        result = testdir.runpytest('--translations', '-vvv', '-r', 's')
        result.stdout.fnmatch_lines([
            "*collected 4*",
            "SKIPPED * aspell dictionary for language hr not found*",
        ])

    def test_python_format_placeholders(self, testdir):
        testdir.makefile(
            '.po',
            """
            msgid ""
            msgstr ""
            "Language: de\\n"

            #: asdf.py:111
            msgid "meeting"
            msgstr "Langer Text %(salkdjfalsdjf)s {kaksjsalkas} Verabredung"
            """
        )
        result = testdir.runpytest('--translations', '-vvv', '-r', 's')
        result.stdout.fnmatch_lines([
            "*collected 4*",
            "*4 passed*",
        ])

    def test_shy_entity(self, testdir):
        testdir.makefile(
            '.po',
            """
            msgid ""
            msgstr ""
            "Language: de\\n"

            #: asdf.py:111
            msgid "meeting"
            msgstr "Ver&shy;ab&shy;redung"
            """
        )
        result = testdir.runpytest('--translations', '-vvv', '-r', 's')
        result.stdout.fnmatch_lines([
            "*collected 4*",
            "*4 passed*",
        ])

    def test_pass(self, testdir):
        testdir.makefile(
            '.po',
            """
            msgid ""
            msgstr ""
            "Language: de\\n"

            #: asdf.py:111
            msgid "meeting"
            msgstr "Verabredung"
            """
        )
        result = testdir.runpytest('--translations', '-vvv', '-r', 's')
        result.stdout.fnmatch_lines([
            "*collected 4*",
            "*4 passed*",
        ])

    def test_fail(self, testdir):
        testdir.makefile(
            '.po',
            """
            msgid ""
            msgstr ""
            "Language: de\\n"

            #: asdf.py:111
            msgid "meeting"
            msgstr "meeeting"
            """
        )
        result = testdir.runpytest('--translations', '-vvv', '-r', 's')
        result.stdout.fnmatch_lines([
            "*collected 4*",
            '*Spell checking failed:*',
            '*msgstr "meeeting"*',
            "*1 failed*",
        ])

    def test_wordlist(self, testdir, monkeypatch):
        testdir.makefile(
            '.po',
            """
            msgid ""
            msgstr ""
            "Language: de\\n"

            #: asdf.py:111
            msgid "meeting"
            msgstr "meeeting"
            """
        )

        test_dir = str(testdir.tmpdir.dirpath())
        wordlist_de = os.path.join(test_dir, 'de')
        assert not os.path.exists(wordlist_de)

        with open(wordlist_de, 'w+') as wl:
            wl.write("meeeting\n")

        monkeypatch.setenv('PYTEST_TRANSLATIONS_PRIVATE_WORD_LIST', test_dir)

        assert os.path.exists(wordlist_de)

        result = testdir.runpytest('--translations', '-vvv')
        result.stdout.fnmatch_lines([
            "*collected 4*",
            "*4 passed*",
        ])



import os
import tempfile

import shutil
from pytest import File, Item

from pytest_translations.config import MARKER_NAME
from pytest_translations.utils import TranslationException, open_po_file, open_mo_file, msgfmt


class MoFileItem(Item, File):
    def __init__(self, fspath, parent):
        super().__init__(fspath, parent)

        if hasattr(self, 'add_marker'):
            self.add_marker(MARKER_NAME)
        else:
            self.keywords[MARKER_NAME] = True

    def runtest(self):
        po_path, _ = os.path.splitext(str(self.fspath))
        po_path += ".po"

        if not os.path.exists(po_path):
            raise TranslationException(
                "corresponding .po file does not exist",
                []
            )

        po_file = open_po_file(po_path)

        temp_dir = tempfile.mkdtemp()

        try:
            test_file = os.path.join(temp_dir, 'test.mo')
            po_file.save_as_mofile(test_file)

            original_parsed = open_mo_file(self.fspath)
            test_parsed = open_mo_file(test_file)

            if len(original_parsed) != len(test_parsed):
                raise TranslationException(
                    'mo-file length does not match po.',
                    []
                )

            diff = []

            for l, r in zip(original_parsed, test_parsed):
                if not (l.msgid == r.msgid and
                        l.msgid_plural == r.msgid_plural and
                        l.msgstr == r.msgstr and
                        l.msgstr_plural == r.msgstr_plural
                        ):
                    diff.append(
                        (l, r)
                    )

            if diff:
                raise TranslationException(
                    'mo-file content does not match po.',
                    diff,
                )

        finally:
            shutil.rmtree(temp_dir)

    def repr_failure(self, excinfo):
        if isinstance(excinfo.value, TranslationException):
            msg, diff = excinfo.value.args

            msg += '\n' + '\n'.join(
                '{0} -> {1}'.format(
                    msgfmt(l),
                    msgfmt(r),
                )
                for l, r in diff
            )

            return msg

        else:
            return super().repr_failure(excinfo)

    def reportinfo(self):
        return (self.fspath, -1, "mo-test")



import os
import re

from pytest import Item, skip

from pytest_translations.config import MARKER_NAME
from pytest_translations.utils import TranslationException

try:
    import enchant
    from enchant.tokenize import (
        get_tokenizer, URLFilter,
        EmailFilter, HTMLChunker
    )
except ImportError:
    enchant = None


def is_enchant_installed():
    return not (enchant is None)


if enchant:
    supported_languages = [
        name.split('_')[0]
        for name, _ in enchant.list_dicts()
    ]
else:
    supported_languages = []


def word_lists():
    PWL = os.getenv('PYTEST_TRANSLATIONS_PRIVATE_WORD_LIST')

    if enchant and PWL and os.path.exists(PWL):
        return {
            fn: os.path.join(PWL, fn)
            for fn in os.listdir(PWL)
            if not os.path.isdir(fn)
        }

    else:
        return {}


class PoSpellCheckingItem(Item):
    def __init__(self, line, language, name, parent):
        self.line = line
        self.language = language

        wl = word_lists()

        private_word_list = wl.get(language, None)

        if enchant and language in supported_languages:
            self.lang_dict = enchant.DictWithPWL(language, pwl=private_word_list)
        else:
            self.lang_dict = None

        super().__init__(name, parent)
        self.add_marker(MARKER_NAME)

    def runtest(self):
        if not enchant:
            skip("enchant is not installed")

        if not self.language:
            skip("no language defined in PO file")

        if self.language not in supported_languages:
            skip(
                "aspell dictionary for language {} not found.".format(self.language)
            )

        text = self.line.msgstr

        # remove python format placeholders, old and new,
        # where it's a problem.
        # 1. replace everything between curly braces
        text = re.sub('{.*?}', '', text)
        # 2. remove everything between %( and )
        text = re.sub('\%\(.*?\)', '', text)
        # 3. remove &shy; html entity
        text = text.replace('&shy;', '')

        tokenizer = get_tokenizer(
            chunkers=[
                HTMLChunker,
            ],
            filters=(
                EmailFilter,
                URLFilter,
            ),
        )
        tokens = tokenizer(text)

        errors = [
            t
            for t, _ in tokens
            if not self.lang_dict.check(t)
            ]

        if errors:
            raise TranslationException(
                "Spell checking failed: {}".format(', '.join(errors)),
                self.line.msgstr,
                [
                    (t, self.lang_dict.suggest(t))
                    for t in errors
                    ]
            )

    def repr_failure(self, excinfo):
        if isinstance(excinfo.value, TranslationException):
            msg, line, wrong = excinfo.value.args
            lines = [msg, "%s" % self.fspath, "msgstr \"%s\"" % line]
            for i in wrong:
                lines.append("%s: %s" % i)

            return "\n".join(lines)

        else:
            return super().repr_failure(excinfo)

    def reportinfo(self):
        return (self.fspath, -1, 'po-spelling')



from pytest import File, Item

from pytest_translations.config import MARKER_NAME
from pytest_translations.po_spelling import PoSpellCheckingItem
from pytest_translations.utils import TranslationException, open_po_file, msgfmt


class PoFile(File):
    def __init__(self, fspath, parent):
        super().__init__(fspath, parent)

        if hasattr(self, 'add_marker'):
            self.add_marker(MARKER_NAME)
        else:
            self.keywords[MARKER_NAME] = True

    def collect(self):
        yield PoUntranslatedItem.from_parent(
            name=self.name,
            parent=self,
        )
        yield PoFuzzyItem.from_parent(
            name=self.name,
            parent=self,
        )
        yield PoObsoleteItem.from_parent(
            name=self.name,
            parent=self,
        )
        try:
            parsed = open_po_file(self.fspath)
        except TranslationException:
            # if the PO file is invalid, we can't do spellchecking.
            # the other tests will fail then, but this here is the collection phase.
            pass
        else:
            language = parsed.metadata.get('Language', '')
            for line in parsed.translated_entries():
                yield PoSpellCheckingItem.from_parent(
                    line=line,
                    language=language,
                    name=self.name,
                    parent=self,
                )


class PoBaseItem(Item):
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.add_marker(MARKER_NAME)

    def repr_failure(self, excinfo):
        if isinstance(excinfo.value, TranslationException):
            msg, wrong = excinfo.value.args

            msg += "\n{0}".format(self.fspath)

            msg += '\n' + '\n'.join(
                msgfmt(i)
                for i in wrong
            )

            return msg

        else:
            return super().repr_failure(excinfo)


class PoUntranslatedItem(PoBaseItem):
    def runtest(self):
        parsed = open_po_file(self.fspath)

        untranslated = parsed.untranslated_entries()

        if not untranslated:
            return

        raise TranslationException(
            'found untranslated entries in file.',
            untranslated,
        )

    def reportinfo(self):
        return (self.fspath, -1, "po-untranslated")


class PoObsoleteItem(PoBaseItem):
    def runtest(self):
        parsed = open_po_file(self.fspath)

        obsolete = parsed.obsolete_entries()

        if not obsolete:
            return

        raise TranslationException(
            'found obsolete entries in file.',
            obsolete,
        )

    def reportinfo(self):
        return (self.fspath, -1, "po-obsolete")


class PoFuzzyItem(PoBaseItem):
    def runtest(self):
        parsed = open_po_file(self.fspath)

        fuzzy = parsed.fuzzy_entries()

        if not fuzzy:
            return

        raise TranslationException(
            'found fuzzy entries in file.',
            fuzzy,
        )

    def reportinfo(self):
        return (self.fspath, -1, "po-fuzzy")
